/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggVorbis 'TREMOR' CODEC SOURCE CODE.   *
 *                                                                  *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS LIBRARY SOURCE IS     *
 * GOVERNED BY A BSD-STYLE SOURCE LICENSE INCLUDED WITH THIS SOURCE *
 * IN 'COPYING'. PLEASE READ THESE TERMS BEFORE DISTRIBUTING.       *
 *                                                                  *
 * THE OggVorbis 'TREMOR' SOURCE CODE IS (C) COPYRIGHT 1994-2002    *
 * BY THE Xiph.Org FOUNDATION http://www.xiph.org/                  *
 *                                                                  *
 ********************************************************************

 function: simple example decoder using vorbisidec

 ********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <kernel.h>
#include <sifrpc.h>
#include <loadfile.h>
#include <tamtypes.h>
#include <debug.h>

#include <audsrv.h>

#include <ivorbiscodec.h>
#include <ivorbisfile.h>

/*Declare audsrv module */
extern unsigned char audsrv[];
extern unsigned int size_audsrv;

/*RAW PCM Buffer */
char pcmout[4096]; 

/*OGG File in memory */
typedef struct _SOggFile {
    char* dataPtr;
    int   dataSize;
    int   dataRead;
}SOggFile;

/*----------------------------------------------------------------------------*/
/*Callback funtions */
/*----------------------------------------------------------------------------*/
size_t VorbisRead(void *ptr, size_t byteSize, size_t sizeToRead, void *datasource) {
	size_t		spaceToEOF;
	size_t		actualSizeToRead;
	SOggFile*	vorbisData;

	vorbisData = (SOggFile*)datasource;
	spaceToEOF = vorbisData->dataSize - vorbisData->dataRead;
	if ((sizeToRead*byteSize) < spaceToEOF)
		actualSizeToRead = (sizeToRead*byteSize);
	else
		actualSizeToRead = spaceToEOF;	
	
	if (actualSizeToRead) {
		memcpy(ptr, (char*)vorbisData->dataPtr + vorbisData->dataRead, actualSizeToRead);
		vorbisData->dataRead += (actualSizeToRead);
	}

	return actualSizeToRead;
}

int VorbisSeek(void *datasource, ogg_int64_t offset, int whence) {
	size_t		spaceToEOF;
	ogg_int64_t	actualOffset;
	SOggFile*	vorbisData;

	vorbisData = (SOggFile*)datasource;

	switch (whence) {
	case SEEK_SET: 
		if (vorbisData->dataSize >= offset)
			actualOffset = offset;
		else
			actualOffset = vorbisData->dataSize;
		vorbisData->dataRead = (int)actualOffset;
		break;
	case SEEK_CUR:
		spaceToEOF = vorbisData->dataSize - vorbisData->dataRead;
		if (offset < spaceToEOF)
			actualOffset = (offset);
		else
			actualOffset = spaceToEOF;	
		vorbisData->dataRead += actualOffset;
		break;
	case SEEK_END:
		vorbisData->dataRead = vorbisData->dataSize+1;
		break;
	default:
		printf("*** ERROR *** Unknown seek command in VorbisSeek\n");
		break;
	};

	return 0;
}

int VorbisClose(void *datasource) {
	return 1;
}

long VorbisTell(void *datasource) {
	SOggFile*	vorbisData;
	vorbisData = (SOggFile*)datasource;
	return vorbisData->dataRead;
}
/*----------------------------------------------------------------------------*/
/*End Callback functions */
/*----------------------------------------------------------------------------*/

void print_logo() {
  init_scr();
  scr_printf("                    .'|_.-\n");
  scr_printf("                  .'  '  /_\n");
  scr_printf("               .-\"     -.   '>\n");
  scr_printf("            .- -. -.    '. /    /|_\n");
  scr_printf("           .-.--.-.       ' >  /  /\n");
  scr_printf("          (o( o( o )       \\_.\"  <\n");
  scr_printf("           '-'-''-'            ) <\n");
  scr_printf("         (       _.-'-.   ._\\.  _\\\n");
  scr_printf("          '----\"/--.__.-) _-  \\|\n");
  scr_printf("                \"V\"\"    \"V\"\n");
  scr_printf("--------------------------------------------------\n");
  scr_printf("OGG VORBIS PLAYER DEMO                            \n");
  scr_printf("--------------------------------------------------\n");
}

int main () {
  OggVorbis_File vf;
  int eof=0;
  int current_section;
  /*eeram buffer variables */
  FILE* iFile;
  int iSize;
  unsigned char* eeram_buffer;
  /*callback variables */
  ov_callbacks vorbisCallbacks;
  SOggFile oggMemoryFile;
  /*audsrv variables */
  int ret, played, err;
  struct audsrv_fmt_t audioformat;

  SifInitRpc(0);

  print_logo();

  printf("oggplayer: kicking IRXs\n");
  ret = SifLoadModule("rom0:LIBSD", 0, NULL);
  printf("libsd loadmodule %d\n", ret);

  printf("oggplayer: loading audsrv\n");
  SifExecModuleBuffer(audsrv, size_audsrv, 0, NULL, &ret);
  printf("audsrv loadmodule %d\n", ret);

#if 0
  /*Modules for USB Access */
  printf("oggplayer: loading usbd\n");
  ret = SifLoadModule("host:usbd.irx", 0, NULL);
  printf("usbd loadmodule %d\n", ret);

  printf("oggplayer: loading usbhdfsd\n");
  ret = SifLoadModule("host:usbhdfsd.irx", 0, NULL);
  printf("usbhdfsd loadmodule %d\n", ret);
#endif

  /*Store callback functions */
  vorbisCallbacks.read_func  = VorbisRead;
  vorbisCallbacks.close_func = VorbisClose;
  vorbisCallbacks.seek_func  = VorbisSeek;
  vorbisCallbacks.tell_func  = VorbisTell;

  /*Open ogg file */
  iFile = fopen("host:/test.ogg","r");
  if (!iFile) {
	  scr_printf("Error opening file\n");
	  return 1;
  }
  printf("fopen ok\n");

  printf("Getting file size\n");
  fseek(iFile, 0, SEEK_END);
  iSize = (int)ftell(iFile);
  fseek(iFile, 0, SEEK_SET);
  printf("got file size ok\n");

  printf("Allocating buffer\n");
  if (!(eeram_buffer = malloc(iSize))) {
    fclose(iFile);
    scr_printf("Couldnt allocate buffer\n");
    printf("Couldnt allocate buffer\n");
    return 1;
  }

  printf("Buffer allocate ok\n");

  /*Copy file into eeram_buffer */
  scr_printf("Buffering OGG File...\n");
  if (!(fread(eeram_buffer, iSize, 1, iFile))) {
    fclose(iFile);
    scr_printf("Read error\n");
    printf("Read error\n");
    return 1;
  }

  printf("Read ok\n");
  fclose(iFile);

  /*Initialize callback structure */
  oggMemoryFile.dataPtr = eeram_buffer;
  oggMemoryFile.dataRead = 0;
  oggMemoryFile.dataSize = iSize;

  if(ov_open_callbacks(&oggMemoryFile, &vf, NULL, 0, vorbisCallbacks) < 0) {
      scr_printf("ERROR: Input does not appear to be an Ogg bitstream.\n");
      printf("Input does not appear to be an Ogg bitstream.\n");
      exit(1);
  }

  /* Throw the comments plus a few lines about the bitstream we're
     decoding */
  {
    scr_printf("Starting Playback...\n");
    scr_printf("---------------------------------------[Song Info]\n");
    char **ptr=ov_comment(&vf,-1)->user_comments;
    vorbis_info *vi=ov_info(&vf,-1);
    while(*ptr){
      scr_printf("%.50s\n", *ptr);
      printf("%s\n",*ptr);
      ++ptr;
    }
    scr_printf("--------------------------------------------------\n");
    printf("\nBitstream is %d channel, %ldHz\n",vi->channels,vi->rate);
    printf("\nDecoded length: %ld samples\n",
	    (long)ov_pcm_total(&vf,-1));
    printf("Encoded by: %s\n\n",ov_comment(&vf,-1)->vendor);
    audioformat.freq = (int)vi->rate;
    audioformat.bits = 16;
    audioformat.channels = vi->channels;
  }

  /*Playback! */
  printf("Start Playback\n");

  ret = audsrv_init();
  if (ret != 0)
  {
    printf("oggplayer: failed to initialize audsrv\n");
    printf("audsrv returned error string: %s\n", audsrv_get_error_string());
    return 1;
  }


  err = audsrv_set_format(&audioformat);
  printf("set format returned %d\n", err);
  printf("audsrv returned error string: %s\n", audsrv_get_error_string());

  audsrv_set_volume(MAX_VOLUME);
  
  printf("starting play loop\n");
  played = 0;

  while(!eof){
    long retz=ov_read(&vf,pcmout,sizeof(pcmout),&current_section);
    if (retz == 0) {
      /* EOF */
      eof=1;
    } else if (retz < 0) {
      /* error in the stream.  Not a problem, just reporting it in
	 case we (the app) cares.  In this case, we don't. */
    } else {
      /* we don't bother dealing with sample rate changes, etc, but
	 you'll have to*/
      audsrv_wait_audio((int)retz);
      audsrv_play_audio(pcmout, (int)retz);
    }
    /* some dots to show progress */
    played++;
    if (played % 128 == 0) {
      printf(".");
    }
  }
  printf("\n");

  /* cleanup */
  printf("oggplayer: stopping audsrv\n");
  audsrv_quit();
  ov_clear(&vf);
  free(eeram_buffer);
    
  scr_printf("Done! ");
  printf("Done!\n");
  return(0);
}
